// This is a personal academic project. Dear PVS-Studio, please check it.
// PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <signal.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <arpa/inet.h>

#include "args.h"
#include "selector.h"
#include "socks5nio.h"
#include "confignio.h"
#include "buffer.h"

#define N(x) (sizeof(x)/sizeof((x)[0]))

static bool done = false;

static void sigterm_handler(const int signal) {
    printf("Signal %d, Cleaning up and exiting\n", signal);
    done = true;
}

static struct socks5args * args;

int main(int argc, char **argv) {
    args = malloc(sizeof(struct socks5args));
    parse_args(argc, argv, args);

    close(STDIN_FILENO);

    const char *err_msg = NULL;
    selector_status ss = SELECTOR_SUCCESS;
    fd_selector selector = NULL;


    // ------
    // TCP
    // ------

    struct sockaddr_in addr;
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    if (inet_pton(AF_INET, args->socks_addr, &addr.sin_addr) <= 0) {
        err_msg = "Incorrect network address";
        goto finally;
    }
    // addr.sin_addr.s_addr = htonl(INADDR_ANY);
    addr.sin_port = htons(args->socks_port);

    const int server = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (server < 0) {
        err_msg = "Unable to create TCP socket";
        goto finally;
    }

    fprintf(stdout, "Listening on TCP port %u\n", args->socks_port);

    setsockopt(server, SOL_SOCKET, SO_REUSEADDR, &(int){ 1 }, sizeof(int));

    if (bind(server, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
        err_msg = "Unable to bind TCP socket";
        goto finally;
    }

    if (listen(server, 20) < 0) {
        err_msg = "Unable to listen";
        goto finally;
    }

    // ------
    // UDP
    // ------

    struct sockaddr_in udp_addr;
    memset(&udp_addr, 0, sizeof(udp_addr));
    udp_addr.sin_family = AF_INET;
    if (inet_pton(AF_INET, args->mng_addr, &udp_addr.sin_addr) <= 0) {
        err_msg = "Incorrect network address";
        goto finally;
    }
    // udp_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    udp_addr.sin_port = htons(args->mng_port);

    const int udp_server = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udp_server < 0) {
        err_msg = "Unable to create UDP socket";
        goto finally;
    }

    fprintf(stdout, "Listening on UDP port %u\n", args->mng_port);

    setsockopt(server, SOL_SOCKET, SO_REUSEADDR, &(int){ 1 }, sizeof(int));

    if (bind(udp_server, (struct sockaddr *) &udp_addr, sizeof(udp_addr)) < 0) {
        err_msg = "Unable to bind UDP socket";
        goto finally;
    }

    if (fcntl(udp_server, F_SETFL, O_NONBLOCK) < 0) {
        err_msg = "Unable to put UDP socket into non-blocking mode";
        goto finally;
    }

    // struct config * config_ret = calloc(1, sizeof(struct config));
    struct config * config_ret = malloc(sizeof(struct config));
    memset(config_ret, 0x00, sizeof(*config_ret));
    buffer_init(&config_ret->read_buffer, N(config_ret->raw_buff_a), config_ret->raw_buff_a);
    buffer_init(&config_ret->write_buffer, N(config_ret->raw_buff_b), config_ret->raw_buff_b);

    signal(SIGTERM, sigterm_handler);
    signal(SIGINT,  sigterm_handler);

    if (selector_fd_set_nio(server) == -1) {
        err_msg = "Getting server socket flags";
        goto finally;
    }
    const struct selector_init conf = {
        .signal = SIGALRM,
        .select_timeout = {
            .tv_sec  = 10,
            .tv_nsec = 0,
        },
    };
    if (0 != selector_init(&conf)) {
        err_msg = "Initializing selector";
        goto finally;
    }

    selector = selector_new(1024);
    if (selector == NULL) {
        err_msg = "Unable to create selector";
        goto finally;
    }

    const struct fd_handler socksv5 = {
        .handle_read = socksv5_passive_accept,
        .handle_write = NULL,
        .handle_close = NULL,
    };
    ss = selector_register(selector, server, &socksv5, OP_READ, NULL);
    if (ss != SELECTOR_SUCCESS) {
        err_msg = "Registering fd";
        goto finally;
    }

    if (selector_fd_set_nio(udp_server) == -1) {
        err_msg = "Getting config socket flags";
        goto finally;
    }

    const struct fd_handler config = {
            .handle_read = config_read,
            .handle_write = config_write,
            .handle_close = NULL,
    };
    ss = selector_register(selector, udp_server, &config, OP_READ, config_ret);
    if (ss != SELECTOR_SUCCESS) {
        err_msg = "Registering fd";
        goto finally;
    }

    for(;!done;) {
        err_msg = NULL;
        ss = selector_select(selector);
        if (ss != SELECTOR_SUCCESS) {
            err_msg = "Serving";
            goto finally;
        }
    }
    if (err_msg == NULL) {
        err_msg = "Closing";
    }

    free(args);
    free(config_ret);

    int ret = 0;
finally:
    if (ss != SELECTOR_SUCCESS) {
        fprintf(stderr, "%s: %s\n", (err_msg == NULL) ? "": err_msg, ss == SELECTOR_IO ? strerror(errno) : selector_error(ss));
        ret = 2;
    } else if (err_msg) {
        perror(err_msg);
        ret = 1;
    }
    if (selector != NULL) {
        selector_destroy(selector);
    }
    selector_close();

    socksv5_pool_destroy();

    if (server >= 0) {
        close(server);
    }
    return ret;
}

int findUser(char * username) {
    for (int i = 0; i < args->nusers; i++) {
        if (!strcmp(args->users[i].name, username))
            return i;
    }
    return -1;
}

int checkPassword(int idx, char * password) {
    return !strcmp(args->users[idx].pass, password);
}
